
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdbool.h>


#define EGG_LENGTH 84
#define EGG_LINES 13


struct Egg {
	char* eggData;
};


struct Egg** eggList = NULL;
size_t numEggs = 0;


const char eggTemplate[] = "\x1b[0;93m"
	                        "     #####\n"
	                        "   ##     ##\n"
	                        "  #         #\n"
	                        " #\x1b[0;91m###########\x1b[93m#\n"
	                        "#             #\n"
	                        "#\x1b[0;94m#############\x1b[93m#\n"
	                        "#             #\n"
	                        "#\x1b[0;91m#############\x1b[93m#\n"
	                        "#             #\n"
	                        " #\x1b[0;94m###########\x1b[93m#\n"
	                        "  #         #\n"
	                        "   ##     ##\n"
	                        "     #####"
	                        "\x1b[0m";



bool is_bad_char(char c)
{
	return !isprint(c) || isspace(c);
}

void error(const char* err)
{
	printf("Error: %s\n", err);
	exit(1);
}

char* build_egg(const char* data, size_t len)
{
	/* create larger buffer to account of utf8 */
	char* eggBuf = (char*)malloc(sizeof(eggTemplate));
	if(!eggBuf) error("OOM");

	memcpy(eggBuf, eggTemplate, sizeof(eggTemplate));

	size_t j = 0;
	for(size_t i = 0; i < sizeof(eggTemplate); ++i)
	{
		if(eggBuf[i] == '#')
		{
			if (j >= len)
				error("Not enough data to generate egg");
			eggBuf[i] = data[j++];
		}
	}
	return eggBuf;
}

char get_random_char()
{
	char c;

	while(is_bad_char(c = rand()));

	return c;
}

char* generate_egg()
{
	char buf[EGG_LENGTH];

	for(size_t i = 0; i < EGG_LENGTH; ++i)
		buf[i] = get_random_char();

	return build_egg(buf, sizeof(buf));
}

bool validate_egg(const char* egg, size_t len)
{
	size_t j = 0;
	size_t i;
	for(i = 0; eggTemplate[i] && j < len; ++i, ++j)
	{
		/* Allow different coloring */
		if(eggTemplate[i] == '\x1b')
		{
			/* Skip color codes */
			for(; i < sizeof(eggTemplate) && eggTemplate[i] != 'm'; ++i);
			++i;
			if(egg[j] == '\x1b')
			{
				for(; j < len && egg[j] != 'm'; ++j);
				++j;
			}
		}

		if(!eggTemplate[i])
			break;

		/* Remove trailing space */
		if(eggTemplate[i] == '\n')
		{
			for(; j < len && egg[j] != '\n'; ++j)
				if(!isspace(egg[j]))
				{
					printf("Newline error\n");
					return false;
				}
		}

		if(eggTemplate[i] == '#')
		{
			if(is_bad_char(egg[j]))
			{
				printf("Bad char detected\n");
				return false;
			}
		}
		else if(eggTemplate[i] != egg[j])
		{
			printf("Mismatch at (%zu,%zu) = ('%c', '%c')\n", i, j, eggTemplate[i], egg[j]);
			return false;
		}
	}

	if(eggTemplate[i])
	{
		printf("Did not reach end of eggTemplate, %zu, '%c'\n", i, eggTemplate[i]);
		return false;
	}

	return true;
}


void parse_egg(char* dest, const char* egg, size_t len)
{
	size_t j = 0;
	for(size_t i = 0; i < len; ++i)
	{
		/* Skip color code */
		if(egg[i] == '\x1b')
		{
			for(; i < len && egg[i] != 'm'; ++i);
			++i;
		}

		if(i >= len)
			break;

		if(!isspace(egg[i]))
		{
			dest[j++] = egg[i];
		}
	}
	dest[j] = 0;
}

void win()
{
	system("get_flag");
}


int get_choice()
{
	char buf[0x10] = {0};
	if(!fgets(buf, sizeof(buf), stdin))
		error("read failed\n");
	return atoi(buf);
}


void add_egg_to_list(struct Egg* egg)
{
	eggList = (struct Egg**)realloc(eggList, ++numEggs * sizeof(struct Egg*));
	if(!eggList) error("Realloc failed!\n");
	eggList[numEggs-1] = egg;
}

bool remove_egg_from_list(size_t idx)
{
	if(idx >= numEggs)
	{
		puts("Egg does not exist!");
		return false;
	}

	struct Egg* egg = eggList[idx];

	free(egg->eggData);
	free(egg);

	memmove(&eggList[idx], &eggList[idx+1], (--numEggs - idx) * sizeof(struct Egg*));

	eggList = (struct Egg**)realloc(eggList, numEggs * sizeof(struct Egg*));

	return true;
}


char* read_egg()
{
	size_t bufSize = sizeof(eggTemplate) * 2;
	size_t bufContent = 0;
	char* buf = (char*)malloc(bufSize);

	size_t numLines = 0;

	while(numLines < EGG_LINES)
	{
		int n = read(STDIN_FILENO, buf + bufContent, bufSize - 1 - bufContent);
		if(n <= 0) error("read failed!\n");


		for(int i = 0; i < n; ++i)
			if(buf[bufContent+i] == '\n')
				numLines++;

		bufContent += n;

		if(bufSize - 1 - bufContent < 0x10)
		{
			buf = (char*)realloc(buf, bufSize *= 2);
			if(!buf) error("OOM");
		}
	}

	buf[bufContent] = 0;

	return buf;
}

void add_egg()
{
	char buf[EGG_LENGTH+1];
	int choice;
	struct Egg* egg = (struct Egg*)malloc(sizeof(struct Egg));

	puts(" 1. Build egg");
	puts(" 2. Parse egg");
	puts(" 3. Generate egg");

	choice = get_choice();


	switch(choice)
	{
		case 1:
			if(!fgets(buf, sizeof(buf), stdin))
				error("Read failed");

			egg->eggData = build_egg(buf, strlen(buf));
			break;

		case 2:
		{
			char* eggstr = read_egg();
			size_t egglen = strlen(eggstr);

			if(!validate_egg(eggstr, egglen))
			{
				puts("Invalid egg!");
				free(eggstr);
				free(egg);
				egg = NULL;
				break;
			}

			parse_egg(buf, eggstr, egglen);
			egg->eggData = build_egg(buf, strlen(buf));
			free(eggstr);
			break;
		}

		case 3:
			egg->eggData = generate_egg();
			break;

		default:
			free(egg);
			egg = NULL;
			puts("Invalid choice");
	}

	if(egg != NULL)
	{
		add_egg_to_list(egg);
		puts("Egg added!");
	}
}


void eat_egg()
{
	printf("Which egg do you wanna eat? ");

	size_t idx = get_choice();

	if(remove_egg_from_list(idx))
		puts("Nom nom nom...");
}


void show_egg()
{
	for(size_t i = 0; i < numEggs; ++i)
		printf("Egg %zu:\n%s\n", i, eggList[i]->eggData);
}


void menu()
{
	puts("-------------");
	puts(" 1. Add egg");
	puts(" 2. Eat egg");
	puts(" 3. Show eggs");
	puts(" 4. Show help");
	puts(" 5. Quit");
	puts("-------------");
}

int main()
{
	setbuf(stdout, NULL);
	setbuf(stdin, NULL);

	int choice;

	bool running = true;

	menu();

	while(running)
	{
		choice = get_choice();
		switch(choice)
		{
			case 1:
				add_egg();
				break;

			case 2:
				eat_egg();
				break;

			case 3:
				show_egg();
				break;

			case 4:
				menu();
				break;

			case 5:
				puts("Bye!");
				running = false;
				break;

			default:
				puts("Invalid choice!");
				menu();
		 }
	}
	return 0;
}
